import json
from google.cloud import storage
from .json_db import Json


def get_db(app_config):
    config = app_config["DB"]
    bucket_name = config["BUCKET_NAME"]
    source_blob_name = config["SOURCE_BLOB_NAME"]
    return GoogleJsonDb(bucket_name, source_blob_name)


def get_blob(bucket_name, source_blob_name):
    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    return bucket.blob(source_blob_name)


def get_data(blob):
    raw_data = blob.download_as_text(client=None)
    return json.loads(raw_data)


class GoogleJsonDb(Json):
    def __init__(self, bucket_name, source_blob_name):
        self.blob = get_blob(bucket_name, source_blob_name)
        data = get_data(self.blob)
        super().__init__(data)

    def save_entry(self, entry):
        data = get_data(self.blob)
        data["data"].append(entry)
        self.blob.upload_from_string(json.dumps(data), content_type='application/json')



import json
import os.path

from platzky.db.json_db import Json


def get_db(config):
    db_path = os.path.abspath(config["DB"]["PATH"])
    return JsonFile(db_path)


class JsonFile(Json):
    def __init__(self, file_path):
        self.data_file_path = file_path
        with open(self.data_file_path) as json_file:
            data = json.load(json_file)
            super().__init__(data)

    def _save_file(self):
        with open(self.data_file_path, 'w') as json_file:
            json.dump(self.data, json_file)

    def add_comment(self, author_name, comment, post_slug):
        super().add_comment(author_name, comment, post_slug)
        self._save_file()



from flask import Flask, request, session, redirect, render_template
from flask_babel import Babel
from flask_minify import Minify

import os
import urllib.parse

from . import config, db_loader
from .blog import blog
from .plugin_loader import plugify
from .seo import seo
from .www_handler import redirect_www_to_nonwww, redirect_nonwww_to_www


def create_app_from_config(config_object):
    engine = create_engine_from_config(config_object)

    blog_blueprint = blog.create_blog_blueprint(db=engine.db,
                                                config=engine.config, babel=engine.babel)
    seo_blueprint = seo.create_seo_blueprint(db=engine.db,
                                             config=engine.config)
    engine.register_blueprint(blog_blueprint)
    engine.register_blueprint(seo_blueprint)
    Minify(app=engine, html=True, js=True, cssless=True)
    return engine


def create_app(config_path):
    absolute_config_path = os.path.join(os.getcwd(), config_path)
    config_object = config.from_file(absolute_config_path)
    return create_app_from_config(config_object)


def create_engine_from_config(config_object):
    config_dict = config_object.asdict()
    db_driver = db_loader.load_db_driver(config_dict["DB"]["TYPE"])
    db = db_driver.get_db(config_dict)
    languages = config_dict["LANGUAGES"]
    domain_langs = config_dict["DOMAIN_TO_LANG"]
    return create_engine(config_dict, db, languages, domain_langs)


def create_engine(config, db, languages, domain_langs):
    app = Flask(__name__)
    app.config.from_mapping(config)

    app.db = db
    app.babel = Babel(app)
    languages = languages
    domain_langs = domain_langs

    @app.before_request
    def handle_www_redirection():
        if app.config["USE_WWW"]:
            return redirect_nonwww_to_www()
        else:
            return redirect_www_to_nonwww()

    @app.babel.localeselector
    def get_locale():
        domain = request.headers['Host']
        lang = domain_langs.get(domain,
                                session.get('language',
                                            request.accept_languages.best_match(languages.keys(), 'en')))
        session['language'] = lang
        return lang

    def get_langs_domain(lang):
        return languages.get(lang).get('domain')

    @app.route('/lang/<string:lang>', methods=["GET"])
    def change_language(lang):
        if new_domain := get_langs_domain(lang):
            return redirect("http://" + new_domain, code=301)
        else:
            session['language'] = lang
            return redirect(request.referrer)

    @app.context_processor
    def utils():
        return {
            "app_name": app.config["APP_NAME"],
            'languages': languages,
            "current_flag": languages[get_locale()]['flag'],
            "current_language": get_locale(),
            "url_link": lambda x: urllib.parse.quote(x, safe=''),
            "menu_items": app.db.get_menu_items()
        }

    @app.errorhandler(404)
    def page_not_found(e):
        return render_template('404.html', title='404'), 404

    return plugify(app)



from platzky.blog.db import DB
import datetime


def get_db(config):
    return Json(config["DATA"])


class Json(DB):
    def __init__(self, data_dict):
        self.data = data_dict

    def get_all_posts(self, lang):
        posts = (filter(lambda x: x["language"] == lang, self.data.get("posts", [])))
        return list(posts)

    def get_post(self, slug):
        post = next(filter(lambda x: x["slug"] == slug, self.data["posts"]), None)
        return post

    def get_page(self, slug):
        post = next(filter(lambda x: x["slug"] == slug, self.data["pages"]), None)
        return post

    def get_menu_items(self):
        post = self.data.get("menu_items", [])
        return post

    def get_posts_by_tag(self, tag, lang):
        posts = filter(lambda x: tag in x["tags"], self.data["posts"])
        return posts

    def get_all_providers(self):
        return self.data["providers"]

    def get_all_questions(self):
        return self.data["questions"]

    def add_comment(self, author_name, comment, post_slug):
        comment = {
            "author": str(author_name),
            "comment": str(comment),
            "date": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        }
        post_index = next(i for i in range(len(self.data["posts"])) if self.data["posts"][i]["slug"] == post_slug)
        self.data["posts"][post_index]["comments"].append(comment)



from flask import redirect
from functools import partial
from gql import gql


def json_get_redirections(self):
    return self.data.get("redirections", {})


def google_get_redirections(self):
    return self.data.get("redirections", {})


def graphql_get_redirections(self):
    redirections = gql(
        """
        query MyQuery{
          redirections(stage: PUBLISHED){
            source
            destination
          }
        }
        """
    )
    return {x['source']:x['destination'] for x in self.client.execute(redirections)['redirections']}


def get_proper_redirections(db_type):
    redirections = {
        "json_file": json_get_redirections,
        "graph_ql": graphql_get_redirections,
        "google_json": google_get_redirections

    }
    return redirections[db_type]


def process(app):
    app.db.get_redirections = get_proper_redirections(app.config["DB"]["TYPE"])
    redirects = app.db.get_redirections(app.db)
    for source, destiny in redirects.items():
        func = partial(redirect, destiny, code=301)
        func.__name__ = f"{source}-{destiny}"
        app.route(rule=source)(func)

    return app



import smtplib
from functools import partial


def send_mail(sender_email, password, smtp_server, port, receiver_email, subject, message):
    full_message = f'From: {sender_email}\nTo: {receiver_email}\nSubject: {subject}\n\n{message}'

    server = smtplib.SMTP_SSL(smtp_server, port)
    server.ehlo()
    server.login(sender_email, password)
    server.sendmail(sender_email, receiver_email, full_message)
    server.close()


def process(app):
    smtp_setup = app.config["SMTP"]
    app.sendmail = partial(send_mail,
                           smtp_setup["ADDRESS"],
                           smtp_setup["PASSWORD"],
                           smtp_setup["SERVER"],
                           smtp_setup["PORT"])
    return app



from platzky.db import json_file_db, graph_ql_db, google_json_db


def load_db_driver(db_type):
    db_type_to_db_loader = {
        "json_file": json_file_db,
        "google_json": google_json_db,
        "graph_ql": graph_ql_db
    }

    return db_type_to_db_loader[db_type]



from flask import request, render_template,\
    make_response, Blueprint, Markup
from platzky.blog import comment_form, post_formatter
from os.path import dirname


def create_blog_blueprint(db, config, babel):
    url_prefix = config["BLOG_PREFIX"]
    blog = Blueprint('blog', __name__, url_prefix=url_prefix, template_folder=f'{dirname(__file__)}/../templates')

    def locale():
        return babel.locale_selector_func()

    @blog.app_template_filter()
    def markdown(text):
        return Markup(text)

    @blog.errorhandler(404)
    def page_not_found(e):
        return render_template('404.html', title='404'), 404

    @blog.route('/', methods=["GET"])
    def index():
        lang = locale()
        return render_template("blog.html", posts=db.get_all_posts(lang))

    @blog.route('/feed', methods=["GET"])
    def get_feed():
        lang = locale()
        response = make_response(render_template("feed.xml", posts=db.get_all_posts(lang)))
        response.headers["Content-Type"] = "application/xml"
        return response

    @blog.route('/<post_slug>', methods=["POST"])
    def post_comment(post_slug):
        comment = request.form.to_dict()
        db.add_comment(post_slug=post_slug, author_name=comment["author_name"],
                       comment=comment["comment"])
        return get_post(post_slug=post_slug)

    @blog.route('/<post_slug>', methods=["GET"])
    def get_post(post_slug):
        if raw_post := db.get_post(post_slug):
            formatted_post = post_formatter.format_post(raw_post)
            return render_template("post.html", post=formatted_post, post_slug=post_slug,
                                   form=comment_form.CommentForm(), comment_sent=request.args.get('comment_sent'))
        else:
            return page_not_found("no such page")

    @blog.route('/page/<path:page_slug>', methods=["GET"])
    def get_page(page_slug):  # TODO refactor to share code with get_post since they are very similar
        if page := db.get_page(page_slug):
            if cover_image := page.get("coverImage"):
                cover_image_url = cover_image["url"]
            else:
                cover_image_url = None
            return render_template("page.html", page=page, cover_image=cover_image_url)
        else:
            return page_not_found("no such page")

    @blog.route('/tag/<path:tag>', methods=["GET"])
    def get_posts_from_tag(tag):
        lang = locale()
        posts = db.get_posts_by_tag(tag, lang)
        return render_template("blog.html", posts=posts, subtitle=f" - tag: {tag}")

    return blog



import os
import sys
from os.path import dirname, abspath
import importlib.util
import pkgutil


def get_selected_not_installed_plugins(enabled_plugins):
    plugins_root_dir = os.path.join(dirname(abspath(__file__)), 'plugins')
    plugins_dirs = set(os.listdir(plugins_root_dir))

    return enabled_plugins - plugins_dirs


def find_plugins(enabled_plugins):
    plugins_dir = os.path.join(dirname(abspath(__file__)), 'plugins')
    plugins = []

    if selected_not_enabled := get_selected_not_installed_plugins(enabled_plugins):
        raise Exception(f"Plugins {selected_not_enabled} has been selected in config, but has not been installed.")

    for plugin_dir in enabled_plugins:
        module_name = plugin_dir.removesuffix('.py')
        spec = importlib.util.spec_from_file_location(module_name,
                                                      os.path.join(plugins_dir, plugin_dir, "entrypoint.py"))
        plugin = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = plugin
        spec.loader.exec_module(plugin)
        plugins.append(plugin)

    for finder, name, ispkg in pkgutil.iter_modules():
        if name.startswith('platzky_'):
            plugins.append(importlib.import_module(name))

    return plugins


def plugify(app):
    plugins = set(app.config["PLUGINS"])
    for plugin in find_plugins(plugins):
        plugin.process(app)
    return app



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['platzky',
 'platzky.blog',
 'platzky.db',
 'platzky.plugins.redirections',
 'platzky.plugins.sendmail',
 'platzky.seo']

package_data = \
{'': ['*'], 'platzky': ['static/*', 'templates/*']}

install_requires = \
['Flask-Babel>=2.0.0,<3.0.0',
 'Flask-Minify>=0.39,<0.40',
 'Flask-WTF>=1.0.1,<2.0.0',
 'Flask>=2.2.2,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'aiohttp>=3.8.3,<4.0.0',
 'google-cloud-storage>=2.5.0,<3.0.0',
 'gql>=3.4.0,<4.0.0',
 'humanize>=4.3.0,<5.0.0']

setup_kwargs = {
    'name': 'platzky',
    'version': '0.1.19',
    'description': 'Another blog in python',
    'long_description': '![Github Actions](https://github.com/platzky/platzky/actions/workflows/tests.yml/badge.svg?event=push&branch=main)\n[![Coverage Status](https://coveralls.io/repos/github/platzky/platzky/badge.svg?branch=main)](https://coveralls.io/github/platzky/platzky?branch=main)\n\n# platzky\n\nBlog engine in python\n\n# How to run?\n\n1. Install platzky with your favorite dependency management tool (`pip install platzky`)\n2. run `flask --app "platzky:create_app(PATH_TO_YOUR_CONFIG_FILE)" run`\n\n## Configuration\n\nFor details check `config.yml.tpl` file.\n\n\n# API\n`platzky.config.from_file(path_to_config)` - creates _platzky_ config from file (see __config.yml.tpl__)\n`platzky.create_app_from_config(config)` - creates _platzky_ application.\n`platzky.sendmail(receiver_email, subject, message)`- sends email from configured account\n',
    'author': 'None',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import os.path
import yaml


def is_db_ok(mapping):
    if 'DB' not in mapping:
        raise Exception("DB not set")
    if 'TYPE' not in mapping['DB']:
        raise Exception("DB type is not set")
    if mapping['DB']['TYPE'] not in ['graph_ql', 'json_file', 'google_json']:
        raise Exception("DB type is not supported")
    return True


class Config():
    def __init__(self, mapping):
        if is_db_ok(mapping):
            self.config = mapping
        else:
            raise Exception("Config is wrong")

    def add_translations_dir(self, absolute_translation_dir):
        self.config["BABEL_TRANSLATION_DIRECTORIES"] += ";" + absolute_translation_dir

    def asdict(self):
        return self.config


def get_config_mapping(base_config):
    default_config = {  # TODO move it to platzky
        "USE_WWW": True,
        "SEO_PREFIX": "/",
        "BLOG_PREFIX": "/",
        "LANGUAGES": {},
        "DOMAIN_TO_LANG": {},
        "PLUGINS": []
    }

    config = default_config | base_config
    babel_format_dir = ";".join(config.get("TRANSLATION_DIRECTORIES", []))
    config["BABEL_TRANSLATION_DIRECTORIES"] = babel_format_dir
    return config


def from_file(absolute_config_path):
    with open(absolute_config_path, "r") as stream:
        file_config = yaml.safe_load(stream)
    file_config["CONFIG_PATH"] = absolute_config_path
    config_from_file = from_mapping(file_config)
    config_directory = os.path.dirname(absolute_config_path)
    for x in ["locales", "locale", "translations"]:
        translation_directory = os.path.join(config_directory, x)
        config_from_file.add_translations_dir(translation_directory)

    path_to_module_locale = os.path.join(os.path.dirname(__file__), "./locale")
    config_from_file.add_translations_dir(path_to_module_locale)

    return config_from_file


def from_mapping(mapping):
    config_dict = get_config_mapping(mapping)
    return Config(config_dict)
